"""Install the package."""

#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
from pathlib import Path
from typing import Final

import cfgtools as cfg
from re_extensions import rsplit, word_wrap

here = Path(__file__).parent

# Load the package's meta-data from pyproject.toml.
f = cfg.read_toml(here / "pyproject.toml")
project = f["project"].asdict()
NAME: Final[str] = project["name"]
SUMMARY: Final[str] = project["description"]
HOMEPAGE: Final[str] = project["urls"]["Repository"]
REQUIRES: Final[list[str]] = project["dependencies"]
SOURCE = "src"
LICENSE = (here / project["license-files"][0]).read_text().partition("\n")[0]

# Import the README and use it as the long-description.
readme_path = here / project["readme"]
if readme_path.exists():
    long_description = "\n" + readme_path.read_text(encoding="utf-8")
else:
    long_description = SUMMARY


def _readme2doc(
    readme: str,
    name: str = NAME,
    requires: list[str] = REQUIRES,
    homepage: str = HOMEPAGE,
    pkg_license: str = LICENSE,
) -> tuple[str, str]:
    doc, rd = "", ""
    for i, s in enumerate(rsplit("\n## ", readme)):
        head = re.search(" .*\n", s).group()[1:-1]
        if i == 0:
            s = re.sub("^\n# .*", f"\n# {name}", s)
        elif head == "Requirements":
            s = re.sub(
                "```txt.*```",
                "```txt\n" + "\n".join(requires) + "\n```",
                s,
                flags=re.DOTALL,
            )
        elif head == "Installation":
            s = re.sub(
                "```sh.*```", f"```sh\n$ pip install {name}\n```", s, flags=re.DOTALL
            )
        elif head == "See Also":
            pypipage = f"https://pypi.org/project/{name}/"
            s = re.sub(
                "### PyPI project\n.*",
                f"### PyPI project\n* {pypipage}",
                re.sub(
                    "### Github repository\n.*",
                    f"### Github repository\n* {homepage}",
                    s,
                ),
            )
        elif head == "License":
            s = f"\n## License\nThis project falls under the {pkg_license}.\n"

        rd += s
        if head not in {"Installation", "Requirements", "History"}:
            doc += s
    doc = re.sub("<!--html-->.*<!--/html-->", "", doc, flags=re.DOTALL)
    return word_wrap(doc, maximum=88) + "\n\n", rd


def _quote(readme: str) -> str:
    if "'''" in readme and '"""' in readme:
        raise ReadmeFormatError("Both \"\"\" and ''' are found in the README")
    if '"""' in readme:
        return f"'''{readme}'''"
    else:
        return f'"""{readme}"""'


class ReadmeFormatError(Exception):
    """Raised when the README has a wrong format."""


if __name__ == "__main__":
    # Import the __init__.py and change the module docstring.
    init_path = here / SOURCE / NAME / "__init__.py"
    module_file = init_path.read_text(encoding="utf-8")
    new_doc, long_description = _readme2doc(long_description)
    module_file = re.sub(
        "^\"\"\".*\"\"\"|^'''.*'''|^", _quote(new_doc), module_file, flags=re.DOTALL
    )
    init_path.write_text(module_file, encoding="utf-8")
    readme_path.write_text(long_description.strip(), encoding="utf-8")
    os.system(f"cd {here} && python -m build")



"""
# loggings
Provides more logging tools.

## Usage
```py
>>> import loggings

>>> loggings.warning("hello")
WARNING:root:hello

>>> loggings.get_logger("here", loggings.DEBUG).debug("hello")
DEBUG:here:hello
```

## See Also
### Github repository
* https://github.com/Chitaoji/loggings/

### PyPI project
* https://pypi.org/project/loggings/

## License
This project falls under the BSD 3-Clause License.

"""

from typing import List

from . import core
from .core import *

__all__: List[str] = []
__all__.extend(core.__all__)



"""
Contains the core of $package: ... , etc.

NOTE: this module is private. All functions and objects are available in the main
`$package` namespace - use that instead.

"""

import logging
import sys
import traceback
from logging import CRITICAL, DEBUG, ERROR, FATAL, INFO, NOTSET, WARN, WARNING
from typing import TYPE_CHECKING, Mapping

if TYPE_CHECKING:
    from logging import _ExcInfoType, _Level


_ANSI_RESET = "\x1b[0m"
_ANSI_LEVEL_COLORS = {
    DEBUG: "\x1b[36m",
    INFO: "\x1b[32m",
    WARNING: "\x1b[33m",
    ERROR: "\x1b[31m",
    CRITICAL: "\x1b[35m",
}


__all__ = [
    "get_logger",
    "CRITICAL",
    "DEBUG",
    "ERROR",
    "FATAL",
    "INFO",
    "NOTSET",
    "WARN",
    "WARNING",
    "critical",
    "debug",
    "error",
    "fatal",
    "info",
    "warn",
    "warning",
    "log",
]


def get_logger(
    name: str | None = None, level: "_Level | None" = None
) -> logging.Logger:
    """
    Return a logger with the specified name, creating it and adding a
    default handler if necessary.

    Parameters
    ----------
    name : str | None, optional
        Logger name, by default None. If not specified, use `__name__`
        of the calling function's module.
    level : _Level | None, optional
        Specifies logging level of the logger.

    Returns
    -------
    logging.Logger
        Logger with the specified name.

    """

    if name is None:
        name = __get_module_name()
    if name == "__main__":
        name = None
    logger = logging.getLogger(name)
    logger.propagate = False
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = __get_formatter(handler)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    if level is not None:
        logger.setLevel(level)
    return logger


class _AnsiColorFormatter(logging.Formatter):
    """Format log output with ANSI colors by level."""

    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)
        color = _ANSI_LEVEL_COLORS.get(record.levelno)
        if color is None:
            return message
        split_at = message.find(record.getMessage())
        if split_at == -1:
            return f"{color}{message}{_ANSI_RESET}"

        prefix = message[:split_at]
        body = message[split_at:]
        return f"{color}{prefix}{_ANSI_RESET}{body}"


def __supports_color(stream: object) -> bool:
    if stream is not None and hasattr(stream, "isatty") and stream.isatty():
        return True

    return "ipykernel" in sys.modules


def __get_formatter(handler: logging.StreamHandler) -> logging.Formatter:
    fmt = "%(levelname)s:%(name)s:%(message)s"
    stream = getattr(handler, "stream", None)
    if __supports_color(stream):
        return _AnsiColorFormatter(fmt)
    return logging.Formatter(fmt)


def critical(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'CRITICAL' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).critical(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def debug(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'DEBUG' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).debug(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def error(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'ERROR' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).error(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def fatal(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'CRITICAL' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).critical(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def info(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'INFO' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).info(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def warn(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'WARNING' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).warning(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def warning(
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with severity 'WARNING' on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).warning(
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def log(
    level: int,
    msg: object,
    *args: object,
    exc_info: "_ExcInfoType" = None,
    line_info: bool = False,
    stack_info: bool = False,
    stacklevel: int = 1,
    extra: Mapping[str, object] | None = None,
) -> None:
    """Log 'msg % args' with the integer severity `level` on the module logger."""
    if line_info:
        msg = __add_line_info(msg, stacklevel)
        stack_info = False
    get_logger(__get_module_name()).log(
        level,
        msg,
        *args,
        exc_info=exc_info,
        stack_info=stack_info,
        stacklevel=stacklevel,
        extra=extra,
    )


def __get_module_name() -> str:
    frame = sys._getframe(2)  # pylint: disable=protected-access
    return frame.f_globals["__name__"]


def __add_line_info(msg: str, stacklevel: int) -> str:
    stack = traceback.extract_stack()[-max(stacklevel + 2, 1)]
    msg = (
        f"{msg}\n  file {stack.filename}, line {stack.lineno}, in {stack.name}"
        f"\n    {stack.line}"
    )
    return msg
